/* rpu.h

   Copyright (c) 2003-2026 HandBrake Team
   This file is part of the HandBrake source code
   Homepage: <http://handbrake.fr/>.
   It may be used under the terms of the GNU General Public License v2.
   For full terms see the file COPYING file or visit http://www.gnu.org/licenses/gpl-2.0.html
 */

#ifndef HANDBRAKE_RPU_H
#define HANDBRAKE_RPU_H

#define RPU_MODE_UPDATE_ACTIVE_AREA  1
#define RPU_MODE_CONVERT_TO_8_1      2

#define RPU_MODE_EMIT_UNSPECT_62_NAL 4
#define RPU_MODE_EMIT_T35_OBU        8

#endif // HANDBRAKE_RPU_H
